<?php declare(strict_types=1);

namespace App\Http\Controllers\Site;

use App\Dto\QuerySettingsDto;
use App\Http\Requests\Site\AccessTokens\IndexRequest;
use App\Http\Requests\Site\AccessTokens\StoreUpdateRequest;
use App\Http\Resources\Site\AccessTokens\Store;
use App\Services\Site\AccessTokenService;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\RedirectResponse;
use Illuminate\Http\Request;
use Illuminate\View\View;

final class AccessTokensController extends Controller
{
    public function __construct(
        private readonly AccessTokenService $accessTokenService
    ) { }

    public function index(IndexRequest $request): View
    {
        $user = $request->user();
        $data = $request->getDto();
        $querySettingsDto = new QuerySettingsDto(
            limit: 20,
            page: $data->getPage(),
            queryWith: []
        );

        $result = $this->accessTokenService->index($querySettingsDto, $user);
        if ($result->isError()) {
            $this->errors($result);
        }

        return view('site.access-tokens.index', $result->getData());
    }

    public function create(Request $request): View
    {
        $user = $request->user();
        $result = $this->accessTokenService->create($user);
        if ($result->isError()) {
            $this->errors($result);
        }

        return view('site.access-tokens.create', $result->getData());
    }

    public function edit(int $id, Request $request): View
    {
        $user = $request->user();
        $result = $this->accessTokenService->edit($id, $user);
        if ($result->isError()) {
            $this->errors($result);
        }

        return view('site.access-tokens.edit', $result->getData());
    }

    public function store(StoreUpdateRequest $request): JsonResponse
    {
        $data = $request->getDto();
        $user = $request->user();
        $result = $this->accessTokenService->store($data, $user);
        if ($result->isError()) {
            return response()->json($result->getData())->setStatusCode($result->getCode());
        }
        return response()->json(new Store($result));
    }

    public function update(int $id, StoreUpdateRequest $request): RedirectResponse
    {
        $data = $request->getDto();
        $user = $request->user();
        $result = $this->accessTokenService->update($id, $data, $user);
        if ($result->isError()) {
            return redirect()->back()->withInput()->withErrors($result->getErrorsOrMessage());
        }

        return redirect()->route('profile.access-tokens.edit', $result->getModel())->withSuccess($result->getMessage());
    }

    public function destroy(int $id, Request $request): RedirectResponse
    {
        $user = $request->user();
        $result = $this->accessTokenService->destroy($id, $user);
        if ($result->isError()) {
            return redirect()->back()->withInput()->withErrors($result->getErrorsOrMessage());
        }

        return redirect()->route('profile.access-tokens.index')->withSuccess($result->getMessage());
    }
}
