<?php declare(strict_types=1);

namespace App\Services\Registry\V2;

use App\Dto\Service\Registry\V2\EventService\DataProcessing;
use App\Dto\Service\Registry\V2\EventService\Event;
use App\Dto\Service\Registry\V2\EventService\ImageDataPull;
use App\Dto\Service\Registry\V2\EventService\ImageDataPush;
use App\Exceptions\Services\Registry\V2\EventService\DataProcessingException;
use App\Jobs\Registry\V2\ProcessEventPull;
use App\Jobs\Registry\V2\ProcessEventPush;
use App\Repositories\RepositoryRepository;
use App\Repositories\UserRepository;
use App\ServiceResults\ServiceResultError;
use App\ServiceResults\ServiceResultSuccess;
use App\Services\Service;
use App\Services\TagRepository\IncrementPullCommand;
use App\Services\TagRepository\UpdateOrCreateCommand;
use Illuminate\Support\Facades\DB;

final class EventService extends Service
{
    public function __construct(
        private readonly UserRepository        $userRepository,
        private readonly RepositoryRepository  $repositoryRepository,
        private readonly ArchitectureCommand   $architectureCommand,
        private readonly UpdateOrCreateCommand $updateOrCreateCommand,
        private readonly IncrementPullCommand  $incrementPullCommand,
    ) { }

    public function dataProcessing(DataProcessing $dataProcessing): ServiceResultError | ServiceResultSuccess
    {
        $events = $dataProcessing->getEvents()->getEvents();

        foreach ($events as $event) {
            /**
             * @var Event $event
             */

            if ($event->getAction() === 'pull') {
                ProcessEventPull::dispatch($event->getTarget())->onQueue('normal');
                continue;
            }

            if ($event->getAction() === 'push') {
                ProcessEventPush::dispatch($event->getTarget())->onQueue('high');
                continue;
            }

            $exception = new DataProcessingException('Unknown action: ' . $event->getAction() . ' target: ' . print_r($event->getTarget(), true));
            \report($exception);
        }

        return $this->ok('OK!');
    }

    public function push(ImageDataPush $data): ServiceResultError | ServiceResultSuccess
    {
        $user = $this->userRepository->getUserByUsername($data->getUsername(), withTrashed: true);
        if (\is_null($user)) {
            return $this->errNotFound(__('Not Found') . ' User: ' . $data->getUsername());
        }

        $repository = $this->repositoryRepository->getRepositoryByName($user, $data->getRepositoryName(), withTrashed: true);
        if (\is_null($repository)) {
            return $this->errNotFound(__('Not Found') . ' Repository: ' . $data->getRepositoryName());
        }
        unset($user);

        $repositoryRegistry = $data->getUsername() . '/' . $data->getRepositoryName();
        $architecture = $this->architectureCommand->execute($repositoryRegistry, $data->getDigest());
        if ($architecture->isError()) {
            return $architecture;
        }
        $architecture = $architecture->getArchitecture();

        try {
            DB::transaction(function () use ($repository, $architecture, $data) {
                $dataTag = [
                    'digest' => $data->getDigest(),
                    'size'   => $data->getSize(),
                ];
                $this->updateOrCreateCommand->execute($repository, $architecture, $data->getTag(), $dataTag);
            });
        } catch (\Throwable $e) {
            report($e);
            return $this->errService($e->getMessage());
        }

        return $this->ok('OK!');
    }

    public function pull(ImageDataPull $data): ServiceResultError | ServiceResultSuccess
    {
        $user = $this->userRepository->getUserByUsername($data->getUsername(), withTrashed: true);
        if (\is_null($user)) {
            return $this->errNotFound(__('Not Found') . ' User: ' . $data->getUsername());
        }

        $repository = $this->repositoryRepository->getRepositoryByName($user, $data->getRepositoryName(), withTrashed: true);
        if (\is_null($repository)) {
            return $this->errNotFound(__('Not Found') . ' Repository: ' . $data->getRepositoryName());
        }
        unset($user);

        try {
            $result = DB::transaction(function () use ($repository, $data) {
                return $this->incrementPullCommand->execute($repository, $data->getTag(), $data->getDigest());
            });
            if ($result->isError()) {
                return $result;
            }
        } catch (\Throwable $e) {
            report($e);
            return $this->errService($e->getMessage());
        }

        return $this->ok('OK!');
    }
}
